import augmentation.augmentation_2d as aug2d
import augmentation.augmentation_3d as aug3d
from torchvision.transforms import Compose


def get_augmenter2d(shape, deformation_scale=0.01, sigma_max=0.1, include_segmentation=False):
    """
    Default augmenter for 2D data
    :param shape: shape of the inputs that the augmenter will receive
    :param deformation_scale: scale of the deformations
    :param sigma_max: maximum noise standard deviation
    :param include_segmentation: flag that specifies whether the second half of the batch are segmentations
    :return: the augmenter
    """
    augmenter = Compose([
        aug2d.ToFloatTensor(cuda=True),
        aug2d.Rotate90(shape),
        aug2d.FlipX(shape, prob=0.5),
        aug2d.FlipY(shape, prob=0.5),
        aug2d.RandomDeformation(shape, sigma=deformation_scale, include_segmentation=include_segmentation),
        aug2d.AddNoise(sigma_max=sigma_max, include_segmentation=include_segmentation)
    ])

    return augmenter


def get_augmenter3d(shape, deformation_scale=0.01, sigma_max=0.1, include_segmentation=False):
    """
    Default augmenter for 3D data
    :param shape: shape of the inputs that the augmenter will receive
    :param deformation_scale: scale of the deformations
    :param sigma_max: maximum noise standard deviation
    :param include_segmentation: flag that specifies whether the second half of the batch are segmentations
    :return: the augmenter
    """
    augmenter = Compose([
        aug3d.ToFloatTensor(cuda=True),
        aug3d.Rotate90(shape),
        aug3d.FlipX(shape, prob=0.5),
        aug3d.FlipY(shape, prob=0.5),
        aug3d.RandomDeformation(shape, sigma=deformation_scale, include_segmentation=include_segmentation),
        aug3d.AddNoise(sigma_max=sigma_max, include_segmentation=include_segmentation)
    ])

    return augmenter



import os

import h5py
import numpy as np
import tifffile as tiff


def read_tif(file, dtype='uint8'):
    """
    Reads tif formatted file and returns the data in it as a numpy array
    :param file: path to the tif file
    :param dtype: data type of the numpy array
    :return: numpy array containing the data
    """

    data = tiff.imread(file).astype(dtype)

    return data


def read_hdf5(file, dtype='uint8', key=None):
    """
    Reads an HDF5 file as a numpy array
    :param file: path to the hdf5 file
    :param dtype: data type of the numpy array
    :param key: key in the hfd5 file that provides access to the data
    :return: numpy array containing the data
    """
    f = h5py.File(file, 'r')
    data = np.array(f.get(key), dtype=dtype)
    f.close()

    return data


def write_tif(file, data):
    """
    Write tif file from a numpy array
    :param file: the file to be written
    :param data: path to the tif file
    """

    tiff.imwrite(file, data)


def imwrite3D(x, dir, prefix='', rescale=False, z_start=0):
    """
    Write a 3D data set to a directory, slice by slice, as tif files
    :param x: 3D data array
    :param dir: directory to write the data to
    :param prefix: prefix of the separate files
    :param rescale: rescale the data if necessary
    :param z_start: starting index
    """

    if not os.path.exists(dir):
        os.mkdir(dir)
    for i in range(z_start, z_start + x.shape[0]):
        if rescale:
            tiff.imsave(dir + '/' + prefix + str(i) + '.tif', (x[i - z_start, :, :] * 255).astype('uint8'))
        else:
            tiff.imsave(dir + '/' + prefix + str(i) + '.tif', (x[i - z_start, :, :]).astype('uint8'))



"""
    This is a script that illustrates training a U-Net
"""

"""
    Necessary libraries
"""
import argparse
import datetime
import os

import torch
import torch.optim as optim
from torch.utils.data import DataLoader

from neuralnets.data.datasets import StronglyLabeledVolumeDataset
from neuralnets.networks.unet import UNet2D
from neuralnets.util.losses import CrossEntropyLoss
from neuralnets.util.preprocessing import get_augmenter2d
from neuralnets.util.validation import validate

"""
    Parse all the arguments
"""
print('[%s] Parsing arguments' % (datetime.datetime.now()))
parser = argparse.ArgumentParser()

# logging parameters
parser.add_argument("--log_dir", help="Logging directory", type=str, default="unet")
parser.add_argument("--print_stats", help="Number of iterations between each time to log training losses",
                    type=int, default=50)

# network parameters
parser.add_argument("--data_dir", help="Data directory", type=str, default="../../../data")
parser.add_argument("--input_size", help="Size of the blocks that propagate through the network",
                    type=str, default="256,256")
parser.add_argument("--fm", help="Number of initial feature maps in the segmentation U-Net", type=int, default=16)
parser.add_argument("--levels", help="Number of levels in the segmentation U-Net (i.e. number of pooling stages)",
                    type=int, default=4)
parser.add_argument("--dropout", help="Dropout", type=float, default=0.0)
parser.add_argument("--norm", help="Normalization in the network (batch or instance)", type=str, default="instance")
parser.add_argument("--activation", help="Non-linear activations in the network", type=str, default="relu")

# optimization parameters
parser.add_argument("--lr", help="Learning rate of the optimization", type=float, default=1e-3)
parser.add_argument("--step_size", help="Number of epochs after which the learning rate should decay",
                    type=int, default=10)
parser.add_argument("--gamma", help="Learning rate decay factor", type=float, default=0.9)
parser.add_argument("--epochs", help="Total number of epochs to train", type=int, default=200)
parser.add_argument("--len_epoch", help="Number of iteration in each epoch", type=int, default=1000)
parser.add_argument("--test_freq", help="Number of epochs between each test stage", type=int, default=1)
parser.add_argument("--train_batch_size", help="Batch size in the training stage", type=int, default=1)
parser.add_argument("--test_batch_size", help="Batch size in the testing stage", type=int, default=1)

args = parser.parse_args()
args.input_size = [int(item) for item in args.input_size.split(',')]
loss_fn = CrossEntropyLoss()

"""
    Setup logging directory
"""
print('[%s] Setting up log directories' % (datetime.datetime.now()))
if not os.path.exists(args.log_dir):
    os.mkdir(args.log_dir)

"""
    Load the data
"""
input_shape = (1, args.input_size[0], args.input_size[1])
print('[%s] Loading data' % (datetime.datetime.now()))
augmenter = get_augmenter2d([args.input_size[0], args.input_size[1]], include_segmentation=True)
train = StronglyLabeledVolumeDataset(os.path.join(args.data_dir, 'EM/EPFL/training.tif'),
                                     os.path.join(args.data_dir, 'EM/EPFL/training_groundtruth.tif'),
                                     input_shape=input_shape, len_epoch=args.len_epoch)
test = StronglyLabeledVolumeDataset(os.path.join(args.data_dir, 'EM/EPFL/testing.tif'),
                                    os.path.join(args.data_dir, 'EM/EPFL/testing_groundtruth.tif'),
                                    input_shape=input_shape, len_epoch=args.len_epoch)
train_loader = DataLoader(train, batch_size=args.train_batch_size)
test_loader = DataLoader(test, batch_size=args.train_batch_size)

"""
    Build the network
"""
print('[%s] Building the network' % (datetime.datetime.now()))
net = UNet2D(feature_maps=args.fm, levels=args.levels, dropout_enc=args.dropout, dropout_dec=args.dropout,
             norm=args.norm, activation=args.activation)

"""
    Setup optimization for training
"""
print('[%s] Setting up optimization for training' % (datetime.datetime.now()))
optimizer = optim.Adam(net.parameters(), lr=args.lr)
scheduler = optim.lr_scheduler.StepLR(optimizer, step_size=args.step_size, gamma=args.gamma)

"""
    Train the network
"""
print('[%s] Starting training' % (datetime.datetime.now()))
net.train_net(train_loader, test_loader, loss_fn, optimizer, args.epochs, scheduler=scheduler,
              augmenter=augmenter, print_stats=args.print_stats, log_dir=args.log_dir)

"""
    Validate the trained network
"""
validate(net, test.data, test.labels, args.input_size, batch_size=args.test_batch_size,
         write_dir=os.path.join(args.write_dir, 'segmentation_final'),
         val_file=os.path.join(args.log_dir, 'validation_final.npy'))
net = torch.load(os.path.join(args.log_dir, 'best_checkpoint.pytorch'))
validate(net, test.data, test.labels, args.input_size, batch_size=args.test_batch_size,
         write_dir=os.path.join(args.write_dir, 'segmentation_best'),
         val_file=os.path.join(args.log_dir, 'validation_best.npy'))

print('[%s] Finished!' % (datetime.datetime.now()))



import torch
import torch.nn as nn
import torch.nn.functional as F


class CrossEntropyLoss(nn.Module):

    def __init__(self, size_average=False):
        """
        Initialization of the cross entropy loss function
        :param size_average: flag that specifies whether to apply size averaging at the end or not
        """

        super(CrossEntropyLoss, self).__init__()

        self.size_average = size_average

    def forward(self, input, target):

        # apply log softmax
        log_p = F.log_softmax(input, dim=1)

        # channels on the last axis
        input_size = input.size()
        for d in range(1, len(input_size) - 1):
            log_p = log_p.transpose(d, d + 1)
        log_p = log_p.contiguous()

        # reshape everything
        log_p = log_p[target[:, 0, ...].unsqueeze(-1).repeat_interleave(input_size[1], dim=-1) >= 0]
        log_p = log_p.view(-1, input_size[1])
        mask = target >= 0
        target = target[mask]

        # compute negative log likelihood
        loss = F.nll_loss(log_p, target)

        # size averaging if necessary
        if self.size_average:
            loss /= mask.data.sum()

        return loss


class MSELoss(nn.Module):

    def forward(self, input, target):
        target_rec = torch.sigmoid(input)
        loss = torch.mean((target - target_rec) ** 2)
        return loss
